/* Copyright 1996, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

#include "backtalk.h"
#include "sysdict.h"
#include "access.h"
#include "str.h"
#include "ps_acl.h"

/* Access files are an alternative to Yapp style 'acl' files.  If a conference
 * directory has an 'access' file in it, we read it and take permissions from
 * that.
 *
 * The contents of a 'access' file might look like this:
 *
 *    registered -> maypost mayresp
 *    f:observer -> -maypost -mayresp
 *    itemauthor -> maykill mayretitle
 *    fw         -> maylock
 *
 * Each line starts with a collection of conditions, followed by a "->"
 * separator, and a list of access types.  A condition is true if the current
 * user is in that set of people.  A condition with a "-" prefix is true if
 * the current user is not in that set of people.  If all conditions in a rule
 * are true, then the access types on the rhs are granted to the current user
 * (or removed from the current user if the access type is prefixed with a "-").
 * If multiple rules set the same access type, then the last one that fired
 * takes effect.
 *
 * Conditions are mostly the same as the ones in 'acl' files.
 *
 * The rights granted by these rules are:
 *
 *  CONTEXT: CONFERENCE
 *     mayread - who may enter the conference and read the items/respones?
 *     maylock - who may change conf access setup (mode, fwlist, acl, access)
 *     mayname - who may retitle the conference
 *     maypost - who may enter new items
 *     maylinkin - who may link in items
 *
 *  CONTEXT: ITEM
 *     mayresp - who may append responses to an item
 *     mayfreeze - who may freeze/thaw an item
 *     maykill - who may kill an item
 *     mayretitle - who may retitle an item
 *     mayretire - who may retire an item
 *     maylinkout - who may an items out
 *
 *  CONTEXT: RESPONSE
 *      mayedit - who may change the text of an item or response
 *      mayerase - who may erase item text or a response
 *      mayhide - who may hide/show item text or a response
 */


/* Access table.  Each row describes an access type, needed to perform
 * particular activities.  Some are conference-level access settings,
 * some are item level, some are response-level.
 */

struct atent {
    char *name;		/* string name of the access level */
    int level;		/* conference, item or resp level setting? ALEV_* */
    int fw;		/* True if fairwitnesses may do this by default */
    } accesstypes[NACCESS] = {

    /* Entries must be in same order as ACCESS_* def's in access.h */

    { "read",		ALEV_CONF, 1 },
    { "post",		ALEV_CONF, 1 },
    { "name",		ALEV_CONF, 0 },
    { "lock",		ALEV_CONF, 0 },
    { "linkin",		ALEV_CONF, 1 },

    { "linkout",	ALEV_ITEM, 1 },
    { "resp",		ALEV_ITEM, 1 },
    { "kill",		ALEV_ITEM, 1 },
    { "retire",		ALEV_ITEM, 1 },
    { "retitle",	ALEV_ITEM, 1 },
    { "freeze", 	ALEV_ITEM, 1 },

    { "hide",		ALEV_RESP, 1 },
    { "erase",		ALEV_RESP, 1 },
    { "edit",		ALEV_RESP, 1 }

    };



/* FIND_ACCESS_TYPE - given a character string describing an access mode,
 * return it's index in the accesstypes table.  "mayread" is equivalent to
 * "read".  Case doesn't matter.  Returns -1 if it is not a known access
 * mode.
 */

int find_access_type(char *name)
{
    int i;

    /* Drop any 'may' prefix */
    if (!strncasecmp("may",name,3))
    	name+= 3;

    for (i= 0; i < NACCESS; i++)
    	if (!strcasecmp(accesstypes[i].name, name))
		return i;

    return -1;
}

/* RULE_FIRES:  Check if the given rule fires. */

int rule_fires (Access rule)
{
    /* If there is no saved procedure, the rule always fires */
    if (rule->proc == NULL) return 1;

    /* Otherwise run the saved expression and return it's value */
    rerun(rule->proc);
    return pop_boolean();
}


/* MAYDO:  Check if you have the given type of access within the given access
 * access array.  For item and response level access types, we should have
 * all the system variables describing that type already set.
 */

int maydo(Access acc, int what)
{
    int enabled;

    if (what > NACCESS) die("Bad access type value");

    /* Cfadm always has access to everything */
    if (bdict(VAR_AMADM)) return 1;

    /* Default settings */
    enabled= (accesstypes[what].fw ? bdict(VAR_AMFW) : 0);

    /* Scan the rules */
    for ( ; acc != NULL; acc= acc->next)
    	if (bit_test(acc->set,what) && rule_fires(acc))
	    enabled= bit_test(acc->val,what);

    return (enabled != 0);
}


/* NEW_RULE:  Create and initialize a new rule. */

Access new_rule()
{
    /* Allocate and initialize rule data structure */
    Access rule= (Access) malloc(sizeof(struct access_rule));
    bit_clear(rule->set, NACCESS);
    rule->proc= NULL;

    return rule;
}


/* FREE_RULE:  Deallocate all memory allocated to a rule */

void free_rule(Access rule)
{
    if (rule->proc != NULL) free_array(rule->proc);
    free(rule);
}


/* RAP_TOKEN:  Append a token to a rule.  Only the line and source of the
 * token are initialized.  A pointer to the new token is returned.
 */

Token *rap_token(Access rule, int *procmem)
{
    Array *a= rule->proc;

    /* Create an array of default size, if needed */
    if (a == NULL)
    {
    	a= rule->proc= (Array *) malloc(sizeof(Array));
	a->lk= 1;
	*procmem= 24;
	a->a= (Token *) malloc(*procmem * sizeof(Token));
	a->sz= 0;
    }

    /* Enlarge the array if needed */
    if (a->sz >= *procmem)
    {
    	*procmem= a->sz + 24;
	a->a= (Token *) realloc(a->a, *procmem * sizeof(Token));
    }

    /* Enlarge the array and return pointer to new last element */
    a->a[a->sz].sfid= -1;
    a->a[a->sz].line= 1;
    return &(a->a[a->sz++]);
}

void rap_symbol(Access rule, int *procmem, int val)
{
    Token *t= rap_token(rule,procmem);
    t->flag= TK_BOUND_SYMBOL;
    t->val= (void *)(intp)val;
}

void rap_integer(Access rule, int *procmem, int val)
{
    Token *t= rap_token(rule,procmem);
    t->flag= TK_INTEGER;
    t->val= (void *)(intp)val;
}

void cond_author(Access rule, int *procmem, int neg, int uidvar, int idvar)
{
    if (neg)
    {
	rap_symbol(rule,procmem,VAR_UID);
	rap_symbol(rule,procmem,uidvar);
	rap_symbol(rule,procmem,FUNC_NE);

	rap_symbol(rule,procmem,VAR_ID);
	rap_symbol(rule,procmem,idvar);
	rap_symbol(rule,procmem,FUNC_NE);
	rap_symbol(rule,procmem,FUNC_OR);

	rap_symbol(rule,procmem,FUNC_CHECK);
    }
    else
    {
	rap_symbol(rule,procmem,VAR_UID);
	rap_symbol(rule,procmem,uidvar);
	rap_symbol(rule,procmem,FUNC_EQ);
	rap_symbol(rule,procmem,FUNC_CHECK);

	rap_symbol(rule,procmem,VAR_ID);
	rap_symbol(rule,procmem,idvar);
	rap_symbol(rule,procmem,FUNC_EQ);
	rap_symbol(rule,procmem,FUNC_CHECK);
    }
}

void cond_boolean(Access rule, int *procmem, int neg, int var)
{
    rap_symbol(rule,procmem,var);
    if (neg) rap_symbol(rule,procmem,FUNC_NOT);
    rap_symbol(rule,procmem,FUNC_CHECK);
}


void cond_itemauthor(Access rule, int *procmem,
		    int neg, char delim, char *arg)
{
    cond_author(rule,procmem,neg, VAR_ITEMAUTHORUID, VAR_ITEMAUTHORID);
}


void cond_respauthor(Access rule, int *procmem,
		    int neg, char delim, char *arg)
{
    cond_author(rule,procmem,neg, VAR_AUTHORUID, VAR_AUTHORID);
}


void cond_itemtext(Access rule, int *procmem,
		    int neg, char delim, char *arg)
{
    rap_symbol(rule,procmem,VAR_RESP);
    rap_integer(rule,procmem,0);
    rap_symbol(rule,procmem,neg ? FUNC_NE : FUNC_EQ);
    rap_symbol(rule,procmem,FUNC_CHECK);

}

void cond_item(Access rule, int *procmem,
		    int neg, char delim, char *arg)
{
    int n= 0;
    if (delim != '\0')
    	die ("'item' condition in 'access' file requires a list of values");

    while (isdigit(*arg))
    {

	rap_symbol(rule,procmem,VAR_ITEM);
	rap_integer(rule,procmem,atoi(arg));
	if (neg)
	{
	    rap_symbol(rule,procmem,FUNC_NE);
	    rap_symbol(rule,procmem,FUNC_CHECK);
	}
	else
	{
	    rap_symbol(rule,procmem,FUNC_EQ);
	    if (n++)
		rap_symbol(rule,procmem,FUNC_OR);
	}

    	arg= firstin(arg,",");
	if (*arg == '\0') break;
	arg++;
    }

    if (!neg && n > 0) rap_symbol(rule,procmem,FUNC_CHECK);
}

void cond_linked(Access rule, int *procmem,
		    int neg, char delim, char *arg)
{
    cond_boolean(rule, procmem, neg, VAR_LINKED);
}

void cond_frozen(Access rule, int *procmem,
		    int neg, char delim, char *arg)
{
    cond_boolean(rule, procmem, neg, VAR_FROZEN);
}

void cond_retired(Access rule, int *procmem,
		    int neg, char delim, char *arg)
{
    cond_boolean(rule, procmem, neg, VAR_RETIRED);
}


struct acent {
    char *name;		/* Name of the condition */
    void (*func)(Access rule, int *procmem, int neg, char delim, char *arg);
			/* Function to generate test */
    } access_cond[] = {

    { "itemauthor", cond_itemauthor },	/* I am author of item */
    { "respauthor", cond_respauthor },	/* I am author of response */
    { "itemtext", cond_itemtext },	/* Response is item text */
    { "item", cond_item },		/* Item number is on given list */
    { "linked", cond_linked },		/* Item is linked */
    { "frozen", cond_frozen },		/* Item is frozen */
    { "retired", cond_retired },	/* Item is retired */

    { NULL, NULL }
    };


/* ACCESS_WORD:  This handles keywords that are used in 'access' files but
 * not in 'acl' files.  All of these are conditional on which item or response
 * is being read, and not always true or always false for a given user/conf.
 *
 * If the word is not an access keyword, then were return false and do nothing
 * else.
 *
 * If the word is an access keyword, then we append a test script to the
 * procedure in the given rule.
 */

int access_word(char *word, Access rule, int *procmem)
{
    int neg, i, len;
    char *arg;
    char delim;

    /* Drop '-' or '+' prefixes, setting flag to indicate if a '-' was there. */
    neg= (*word == '-');
    if (neg || *word == '+') word++;

    /* Find argument delimiter, if any */
    arg= firstin(word,"=:");
    delim= *arg;
    *arg= '\0';


    for (i= 0; access_cond[i].name != NULL; i++)
	if (!strcmp(access_cond[i].name, word))
	{
	    (*access_cond[i].func)(rule,procmem,neg,delim,arg+1);
	    *arg= delim;
	    return 1;
	}

    *arg= delim;
    return 0;
}


/* READ_ACCESS:  Read the access file from the conference with the given
 * directory, and return an access array describing the current user's access
 * to that conference.  Returns NULL if there is no 'access' file.
 */

Access read_access(char *dir)
{
    char fn[BFSZ];
    char word[BFSZ+1];
    int ch, wordlen, line= 1;
    FILE *fp;
    Access ac= NULL, rule, acend;
    int procmem;
    int state= 0;
    int actype;

    /* Open the file */
    snprintf(fn, BFSZ, "%s/access", dir);
    if ((fp= fopen(fn,"r")) == NULL)
    	return NULL;

    /* Parse the file */
    while ((ch= getc(fp))!= EOF)
    {
    	switch (state)
	{
	case 0: /* First time starting new line */
		state= 1;

		/* drops through to next case */

	case 1: /* Still starting new line */

	    if (ch == '#')
	    	state= 9;	/* skip to next line */
	    else if (!isspace(ch))
	    {
		wordlen= 1;
		word[0]= ch;
		rule= new_rule();
		procmem= 0;	/* amount of memory allocated to rule.proc */
	    	state= 2;	/* Read a condition word */
	    }
	    break;
   	
	case 2: /* Read a condition word */
	    if (wordlen == 1 && word[0] == '-' && ch == '>')
	    {
	    	/* Reached end of conditions.  They must have all been true,
		 * or we wouldn't have got this far.  Go set the flags.  */
		wordlen= 0;
		state= 3;
	    }
	    if (isspace(ch))
	    {
	    	/* Found end of condition word.  Check if it is true. */
		word[wordlen]= '\0';
		if (!access_word(word,rule,&procmem) &&
		    !acl_word(word, 'r', fn, line))
		{
		    /* word is a acl word and is false for current user */
		    free_rule(rule);
		    rule= NULL;
		    state= 9;	/* skip rest of line */
		}

		/* Get ready to read next condition */
		wordlen= 0;
	    }
	    else if (wordlen > BFSZ)
	    	die("Condition too long on line %d of %d\n",line,fn);
	    else
		word[wordlen++]= ch;
	    break;

	case 3: /* Read in an access type to set */
	    if (isspace(ch))
	    {
		char *w= word;
		int val= 1;

		if (*w == '+') w++;
		if (*w == '-') { w++; val= 0;}
		if ((actype= find_access_type(w)) < 0)
		    die("Unknown access type \"%s\" on line %d of %s",w,line,fn);

		bit_set(rule->set,actype);
		if (val)
		{
		    bit_set(rule->val,actype);
		}
		else
		{
		    bit_clear(rule->val,actype);
		}

		if (ch == '\n')
		{
		    /* Append finished rule to list of rules */
		    if (ac == NULL)
		    	ac= rule;
		    else
		    	acend->next= rule;
		    acend= rule;
		    rule->next= NULL;

		    state= 0;	/* Start next line */
		}
	    }

	case 9: /* Skip to next line */
	    if (ch == '\n') state= 0;
	    break;
	}

       	if (ch == '\n') line++;
    }
    return ac;
}
