/* Copyright 1996, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

/* USER INFORMATION
 *
 * This file contains routines to access information about the user.  Mostly
 * it calls the lower level routines in udb.c, which in turn call whichever
 * authentication (auth_*.c) and identification (ident_*.c) modules are
 * linked in.
 */

#include "backtalk.h"

#include "btpasswd.h"
#include "user.h"
#include "udb.h"
#include "group.h"

#ifdef DMALLOC
#include "dmalloc.h"
#endif


struct btpasswd *upw= NULL;	/* NULL means user is anonymous */
char *new_fname= NULL;		/* If fullname has changed, this is it */


/* GETUSER - Identify the user and save his database informtion into the
 * upw variable.  On failure, we set upw to NULL.  We return a error code:
 *   0 - success
 *   1 - failure - no authentication - should treat as anonymous user
 *   2 - failure - authentication expired - should re-authenticate
 * This must be called before any call to user_fname(), user_id(), user_dir(),
 * user_uid() or user_gid() and before calls to func_changename(),
 * func_changepass() or func_removeuser() which apply to the current user.
 *
 * This supports several different methods of identifying the user:
 *   ID_REMOTE_USER - read user ID from 'REMOTE_USER' env variable.
 *   ID_SESSION - get session ID from cookie and look up user in session db.
 *   ID_GETUID - person running backtalk is the user.
 * More than one of these can be turned on, in which case they are attempted
 * in the order listed.  If the global variable direct_exec is turned on,
 * we won't do the first two.  If we do use ID_GETUID, the global direct_exec
 * gets turned on.
 *
 * If secure is true, then we don't trust environment variables unless we
 * are running as either HTTPD or root.
 */

int getuser(int secure)
{
    char *nam;
    uid_t realuid;
    int rc= 1;
#ifdef ID_SESSION
    char bf[MAX_LOGIN_LEN+1];
#endif

    realuid= getuid();
    upw= NULL;		/* Yes, this is needed! */

#if defined(ID_REMOTE_USER) || defined(ID_SESSION)
    /* Try getting ID from environment, if environment is trustworthy */
    if ((!secure || realuid == SERVER_UID || realuid == 0) && !direct_exec)
    {
#ifdef ID_REMOTE_USER
	/* Try getting ID from REMOTE_USER variable */
	nam= getenv("REMOTE_USER");
	if (nam == NULL)
#endif
	{
#ifdef ID_SESSION
	    /* Try getting ID from session cookie */
	    if ((rc= session_auth(nam= bf)) != 0)
#endif
		/* No user */
		return rc;
	}

	/* Get more user info */
	upw= dup_passwd(getdbnam(nam));
#ifdef AUTO_CREATE_DIR
	/* Create him a home directory if he doesn't have one */
	if (upw != NULL && access(upw->dir,F_OK))
	{
	    char bf[BFSZ];
	    mk_directory(bf, nam);
	}
#endif
	return (upw == NULL);
    }
#endif /* ID_REMOTE_USER || ID_SESSION */

    /* Running 'backtalk -d' or not running secure Backtalk as HTTPD */

#ifdef ID_GETUID
    /* Try getting user based on UID number */
#ifdef UID_MAP_FILE
    /* So obsolete we never even tested it */
    if ((nam= mapuid(realuid)) != NULL)
	upw= dup_passwd(getdbnam(nam));
#else
    upw= dup_passwd(getdbuid(realuid));
#endif /* UID_MAP_FILE */
    direct_exec= TRUE;
#endif /* ID_GETUID */

#ifdef AUTO_CREATE_DIR
    /* Create him a home directory if he doesn't have one */
    if (upw != NULL && access(upw->dir,F_OK))
    {
	char bf[BFSZ];
	mk_directory(bf, nam);
    }
#endif

    return (upw == NULL);
}



/* USER_FNAME - Return full name of the user.  Returns "" if the user is
 * anonymous.
 */

char *user_fname()
{
    return (new_fname == NULL ?
		(upw == NULL ?  "" : upw->gecos) : new_fname);
}


/* USER_ID - Return login id of the user.  Returns "" if the user is anonymous.
 */

char *user_id()
{
    return (upw == NULL ? "" : upw->name);
}


/* USER_DIR - Return home directory of the user.  Returns "" if the user is
 * anonymous.
 */

char *user_dir()
{
    return (upw == NULL ? "" : upw->dir);
}


/* USER_UID - Return uid of the user.  Returns -1 if he is anonymous.
 */

int user_uid()
{
    return (upw == NULL ? -1 : upw->uid);
}


/* USER_GID - Return primary group id of the user.  Returns -1 if he is
 * anonymous.
 */

int user_gid()
{
    return (upw == NULL ? -1 : upw->gid);
}

/* USER_GNAME - Return the primary group name for the user.  Returns NULL if
 * he is anonymous.
 */

char *user_gname()
{
    static char *gname= NULL;
    int gid= user_gid();

    if (gname != NULL) return gname;
    if (gid == -1) return NULL;
    return gname= groupname(gid);
}


/* USER_CFADM - Return true if the user is in the CFADM group  Return value
 * if 1 if it is our primary group number, 2 if it is a secondary group,
 * and 0 if it is neither.
 */

int user_cfadm()
{
    if (upw == NULL) return 0;
#ifdef CFADM_GID
    return ingroupno(CFADM_GID,upw->gid,upw->name);
#else
#ifdef CFADM_GROUP
    return ingroupname(CFADM_GROUP,upw->gid,upw->name);
#else
    return 0;
#endif
#endif
}


/* USER_GRADM - Return true if the user is in the GRADM group.  Return
 * value is 1 if it is our primary group number, 2 if it is a secondary group,
 * and 0 if it is neither.
 */

int user_gradm()
{
    if (upw == NULL) return 0;
#ifdef GRADM_GID
    return ingroupno(GRADM_GID,upw->gid,upw->name);
#else
#ifdef GRADM_GROUP
    return ingroupname(GRADM_GROUP,upw->gid,upw->name);
#else
    return 0;
#endif
#endif
}


/* FREE_USER - Deallocate memory for user.  Used for memory debugging only.
 */

#ifdef CLEANUP
void free_user()
{
    free_passwd(upw);
    upw= NULL;
}
#endif
